# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Type annotation helpers.
"""
import os
from typing import Dict, List, Sequence, Tuple, TypeVar, Union

__all__ = [
    'MaybeTuple',
    'MaybeList',
    'MaybeSeq',
    'MaybeDict',
    'PathLike',
]

T = TypeVar('T')
MaybeTuple = Union[T, Tuple[T, ...]]
MaybeList = Union[T, List[T]]
MaybeSeq = Union[T, Sequence[T]]
MaybeDict = Union[T, Dict[str, T]]
PathLike = TypeVar('PathLike', str, os.PathLike)
